(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: main.ml,v 1.8 2004/10/05 19:32:58 julia Exp $ *)

(*open Devilast*)
open Lexing 
open Format
open Event_lexer	(* to get cr_index and lineno *)
open List

let input_name = ref ""	(* no filename by default *)
let nocode = ref false
let output_name = ref "out" (* default output name *)

exception  Err of string ;;


let fatal_error s = Printf.eprintf "%s: %s\n" 
    (Filename.basename Sys.argv.(0)) s;
  exit 2


let version = 
  "This Software is (C) copyright INRIA - 2001.\n\
    Alpha 0.0.1 (12/10/2001)\n"


let print_version () =
  print_string version;
  exit 0


let print_pos fname line colfrom colto = 
  sprintf "File \"%s\", line %d, characters %d-%d:\n"
    fname line colfrom colto


let go ch input_name output_name generate_code = 
  let lexbuf = Lexing.from_channel ch in
  Printf.printf "parsing\n"; flush stdout;
  let ast =
    try Event_parser.event_spec Event_lexer.token lexbuf
    with x -> 
      (Parser_error.report_error lexbuf Format.err_formatter x;
       exit 2) in
  Error.init_error();
  Printf.printf "compiling\n"; flush stdout;
  Objects.policy_blocked_types := true;
  let (init_events,system_events,application_events,
       top_level_events,scheduler_events,
       (unmatched_transitions,event_types,block_collapsed_event_types,
	init_automaton,automaton),
       (interface_types,block_collapsed_interface_types),vs_interface_types)
      = Compile_events.compile_config ast in
  Printf.printf "starting VS types\n"; flush stdout;
  Objects.schedtype := Objects.VIRTSCHED;
  let sched_state_effects = Ps2vs2.ps2vs block_collapsed_event_types in
  Objects.schedtype := Objects.PROCSCHED;
  Printf.printf "done with VS types\n"; flush stdout;
  Evpp.pretty_print output_name init_events system_events
    application_events top_level_events scheduler_events
    event_types interface_types vs_interface_types
    init_automaton automaton sched_state_effects;
  Objects.schedtype := Objects.VIRTSCHED;
  Evpp.pretty_print (Printf.sprintf "%s_bc" output_name)
    init_events system_events
    application_events top_level_events scheduler_events
    block_collapsed_event_types block_collapsed_interface_types
    vs_interface_types
    init_automaton automaton sched_state_effects;
  Objects.schedtype := Objects.PROCSCHED;
  Ev2tex.pretty_print (Printf.sprintf "%s.tex" output_name)
    event_types interface_types vs_interface_types
    unmatched_transitions init_automaton automaton [];
  Objects.schedtype := Objects.VIRTSCHED;
  Ev2tex.pretty_print (Printf.sprintf "%s_bc.tex" output_name)
    block_collapsed_event_types block_collapsed_interface_types
    vs_interface_types
    unmatched_transitions init_automaton automaton sched_state_effects;
  Objects.schedtype := Objects.PROCSCHED;
  Error.check_errors("verification")

(* Parsing of command-line arguments *)
	
exception Found_program_name      

let set_output s =  output_name := s

let anonymous s =
  input_name := s; raise Found_program_name

let speclist = 
  [ "-o", Arg.String set_output,
    "<dirname>  Set the name of the output directory";

    "-version", Arg.Unit print_version, 
    "Display version info"
  ]


let usage =
  Printf.sprintf "Usage: %s [options] <filename> \nOptions are:" 
    (Filename.basename Sys.argv.(0))


let main () =
  let ch = 
    try
      Arg.parse speclist anonymous usage;
      fatal_error "No input file"
    with 
      Found_program_name -> open_in !input_name in
  Misc.init (Filename.basename !input_name);
  go ch !input_name !output_name (not !nocode)
    

let _ = 
  try
    Unix.handle_unix_error main ()
  with 
    Sys_error(s)     -> fatal_error s
  | Failure(s)       -> fatal_error s
  | Error.Error(s)   -> fatal_error s
