(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

%{
(*************************************************************************)
(*                                                                       *)
(*                               Bossa                                   *)
(*                                                                       *)
(*                          , projet Compose, INRIA Rennes               *)
(*                                                                       *)
(*   Copyright (C) 2001 INRIA. All rights reserved.                      *)
(*   Institut National de Recherche en Informatique et en Automatique.   *)
(*                                                                       *)
(*************************************************************************)

(* $Id: module_bossa_parser.mly,v 1.8 2004/11/11 16:15:00 julia Exp $ *)

module B = Objects

%}
%token  MODULE VSMODULE MODULES UNSHARED REQUIRES GLOBAL

%token  SCHEDULER TYPE ENUM PROCESS CONST ORDERING LOWEST HIGHEST KEY TRACE
%token  STATES READY RUNNING BLOCKED TERMINATED NOWHERE QUEUE FIFO LIFO 
%token  HANDLER EVENT ON INTERFACE TRANSITION DEFAULT HIGHRES LOWRES
%token  BEFORE AFTER READS OVERRIDEABLE
%token  INT BOOL TIME CYCLES VOID IF ELSE FOREACH IN SWITCH RETURN TIMER PORT
%token  HEAD TAIL FOREACHINC FOREACHDEC
%token  CASE TRUE FALSE AND OR DOT LPAREN RPAREN LCURLY RCURLY BREAK
%token  DOT_DOT LBRACK RBRACK
%token  EQUAL_EQUAL NEQ COMMA SEMICOLON COLON EQL PLUS
%token  MINUS TIMES DIV GT LT COMPLEMENT LEQ GEQ EQUAL
%token  RARROW  BITOR BITAND PLUS_PLUS MINUS_MINUS SELECT EMPTY NEXT SRCONSCHED
%token  SCHEDULEROF
%token  PLUS_EQUAL MINUS_EQUAL TIMES_EQUAL DIV_EQUAL MOD_EQUAL
%token  BITAND_EQUAL BITOR_EQUAL LSHIFT_EQUAL RSHIFT_EQUAL MOD NOT
%token  LSHIFT RSHIFT ERROR
%token  EVENTS TEST EXPRESSIONS
%token  FWD VIRTSCHEDULER DEFER PUBLIC PRIVATE QUESTION SYSTEM STRUCT
%token  ADMIT ADMIT_CRIT ADMIT_ATT ADMIT_DET
%token EOF
%token <int> INTEGER
%token <string> IDENT
%token <string> ENUM_TYPE
%token <string> RANGE_TYPE
%token <string> FN_NAME
%token <string> STATE_NAME
%token <string> MODULE_NAME
%token <string> STRING

%start module_spec
%type <(Mast.super_sched * Objects.sched) * Objects.location> module_spec
/*%type <unit> module_spec*/

/* specify operator precedence to allow simple single production rule for expressions */
%left COMMA                 /* lowest precedence */
%left EQL PLUS_EQUAL MINUS_EQUAL TIMES_EQUAL DIV_EQUAL MOD_EQUAL 
BITAND_EQUAL BITOR_EQUAL LSHIFT_EQUAL RSHIFT_EQUAL RARROW
%left OR
%left AND 
%left BITOR
%left BITAND
%left EQUAL_EQUAL NEQ
%left GT LT GEQ LEQ
%left LSHIFT RSHIFT
%left PLUS MINUS
%left TIMES DIV MOD IN
%right NOT COMPLEMENT 
%left PLUS_PLUS MINUS_MINUS
%left DOT
%nonassoc UMINUS                /* highest precedence */

/* start syntax definition */
%%
module_spec: 
    x_setpos is_default_hr SCHEDULER x_setproc IDENT EQL LCURLY scheddef RCURLY
    EOF
      {
      let (s,(u,g,p,o,a,h,f)) = $8
      in ((Mast.SCHED(Mast.mkSCHED_MODULE($5,$2,s,u,g,p,o,a,h,f,
					  Module_misc.getpos 10)),
	   B.PROCSCHED),
	  Objects.SCHEDULER_NAME (B.mkId $5))
      } 

|    x_setpos is_default_hr VIRTSCHEDULER x_setvirt IDENT EQL LCURLY vscheddef
    RCURLY EOF
      {
      let (s,(u,g,p,o,a,h,f)) = $8
      in ((Mast.SCHED(Mast.mkSCHED_MODULE($5,$2,s,u,g,p,o,a,h,f,
					  Module_misc.getpos 10)),
	   B.VIRTSCHED),
	  Objects.SCHEDULER_NAME (B.mkId $5))
      } 

|   x_setpos MODULE x_setproc x_setmodule IDENT parameters_modules
    LCURLY moduledef RCURLY EOF
	{Module_progtypes.add_module_name $5;
        let ((c,t,p,g,oc,adm,_),(h,i,a)) = $8
        in ((Mast.MODL(Mast.mkMODULE($5,$6,c,t,g,p,oc,adm,h,i,a,
				     Module_misc.getpos 10)),
	     B.PROCSCHED),
	    Objects.MODULE_NAME (B.mkId $5))
        }

|   x_setpos VSMODULE x_setvirt x_setmodule IDENT parameters_modules
    LCURLY vmoduledef RCURLY EOF
	{Module_progtypes.add_module_name $5;
        let ((c,t,p,g,oc,adm,_),(h,i,a)) = $8
        in ((Mast.MODL(Mast.mkMODULE($5,$6,c,t,g,p,oc,adm,h,i,a,
				     Module_misc.getpos 10)),
	     B.VIRTSCHED),
	    Objects.MODULE_NAME (B.mkId $5))
        }
    ;

/* ----------------------------------------------------------------- */
/* high_res/low_res and default in any order */

is_default_hr:
    DEFAULT is_hr                       { true }
  | HIGHRES is_default
      { if (not(!Objects.high_res))
      then
	raise
	  (Error.Error
	     "high-res scheduler must be compiled with the option -hrt");
      $2}
  | LOWRES is_default
      { if (!Objects.high_res)
      then
	raise 
	  (Error.Error
	     "low-res scheduler may not be compiled with the option -hrt");
	$2}
  |                                     { false }
  ;

is_hr:
    HIGHRES
      { if (not(!Objects.high_res))
      then
	raise
	  (Error.Error
	     "high-res scheduler must be compiled with the option -hrt") }
  | LOWRES
    { if (!Objects.high_res)
    then
      raise 
	(Error.Error
	   "low-res scheduler may not be compiled with the option -hrt") }
  |                                     { () }
  ;

is_default:
    DEFAULT                             {true}
  |                                     {false}
  ;

/* ----------------------------------------------------------------- */
/* main scheduler code */

scheddef:
    sched_decl sched_body               {($1,$2)}
    ;

vscheddef:
    sched_decl vsched_body              {($1,$2)}
    ;

sched_decl:
    statedef {$1}
    ;
    
sched_body:
    usesdef withdef 
      {let (g,p,o,a,h,f) = $2 in
       ($1,g,p,o,a,h,f)
      }
    ; 
   
vsched_body:
    usesdef vwithdef 
      {let (g,p,o,a,h,f) = $2 in
       ($1,g,p,o,a,h,f)
      }
    ; 
   
usesdef:
    MODULES LCURLY uses_body RCURLY {$3}
  | MODULES LCURLY RCURLY {[]}
    ;
    
uses_body:
    ident_body {[$1]}
    | ident_body COMMA uses_body {$1::$3}
    ;

ident_body:
    x_setpos IDENT LPAREN sd_ident_body RPAREN 
      { Module_progtypes.add_module_name $2;
        Mast.mkUSES($2,$4,Module_misc.getpos 5)
      }
    ;  
    
sd_ident_body:
                                     {[]}
    | STATE_NAME                     {[$1]}
    | STATE_NAME COMMA sd_ident_body {$1::$3}
    ;

withdef:
    sd_globaldef sd_processdef sd_orderdef sd_admitdef sd_handlerdef
    sd_functiondef
    {($1,$2,$3,$4,$5,$6)}
    ;
    
vwithdef:
    sd_globaldef sd_schedulerdef sd_orderdef sd_admitdef sd_handlerdef
    sd_functiondef
    {($1,$2,$3,$4,$5,$6)}
    ;
    
sd_globaldef:
    GLOBAL LCURLY sd_proc_body RCURLY {$3}
    | {[]}
    ;

sd_processdef:
    PROCESS LCURLY sd_proc_body RCURLY {$3}
    | {[]}
    ;

sd_schedulerdef:
    SCHEDULER LCURLY sd_proc_body RCURLY {$3}
    | {[]}
    ;
   
sd_proc_body:
    sd_proc_def {[$1]}
    | sd_proc_def COMMA sd_proc_body {$1::$3}
    ;

sd_proc_def:
    MODULE_NAME DOT /*smth*/IDENT x_setpos READS MODULE_NAME DOT /*smth*/IDENT 
    {Mast.mkREQUIRESDEF($1,$3,$6,$8,Module_misc.getpos 5)}
    ;
/*    
smth:
    IDENT {$1}
    | IDENT DOT smth {}
    ;
*/
/* ------------------------------------------------------- */

sd_handlerdef:
    HANDLER LCURLY sd_handler_body RCURLY {$3}
    | {[]}
    ;

sd_handler_body:
    th_handler_body SEMICOLON {[$1]}
    | th_handler_body SEMICOLON sd_handler_body {$1::$3}
    ;

th_handler_body:
    x_setpos th_event_type COLON sd_event_type
      {let ln = Module_misc.getpos 4 in
      Mast.mkHANDLERDEF($2,List.map (function et -> et $2) $4, ln)}
    ;
    
th_event_type:
    x_setpos event_type_ {$2(Module_misc.getpos 2)}
    ;
    
sd_event_type:
    line_type {[$1]}
    | line_type COMMA sd_event_type {$1::$3}
    ;
    
line_type:
    x_setpos MODULE_NAME
      {function event -> Mast.mkEVENT_BODY($2,event, Module_misc.getpos 2)}

/* ------------------------------------------------------- */

sd_functiondef:
    INTERFACE LCURLY sd_function_body RCURLY {$3}
    | {[]}
    ;

sd_function_body:
    th_function_body SEMICOLON {[$1]}
    | th_function_body SEMICOLON sd_function_body {$1::$3}
    ;

th_function_body:
    x_setpos IDENT COLON sd_function_type
      {Mast.mkFUNCTIONDEF($2,List.map (function f -> f $2) $4,
			  Module_misc.getpos 4)}
    ;
    
sd_function_type:
    fn_line_type {[$1]}
    | fn_line_type COMMA sd_function_type {$1::$3}
    ;
    
fn_line_type:
    x_setpos MODULE_NAME
      {function nm -> Mast.mkFUNCTION_BODY($2,nm, Module_misc.getpos 2)}

/* ------------------------------------------------------- */

moduledef:
    module_decl module_body 		{($1,$2)}
     ;

vmoduledef:
    vmodule_decl module_body 		{($1,$2)}
     ;

module_decl:
    constdef typedef_decl processdef globaldef orderdef admission
    tracedef
    {($1,$2,$3,$4,$5,$6,$7)}
    ;

vmodule_decl:
    constdef typedef_decl schedtypedef globaldef orderdef admission
    tracedef
    {($1,$2,$3,$4,$5,$6,$7)}
    ;

module_body:
    handlerdef interfacedef aspectdef 		{($1,$2,$3)}
    ;

typedef_decl:
    enum_def typedef_decl              {$1::$2}
  | range_def typedef_decl             {$1::$2}
  |                                    {[]}
  ;

enum_def:
    x_setpos TYPE IDENT EQL ENUM LCURLY enum_body RCURLY SEMICOLON 
    {Module_progtypes.add_enum_type $3;
      Mast.mkENUMDEF($3, $7, Module_misc.getpos 9)}

enum_body:
   IDENT  				{[$1]}
   | IDENT COMMA enum_body        	{$1::$3}
   ;

range_def:
    x_setpos TYPE IDENT EQL LBRACK expr DOT_DOT expr RBRACK SEMICOLON
    {Module_progtypes.add_range_type $3;
      Mast.mkRANGEDEF($3, $6, $8, Module_misc.getpos 10)}

processdef:
   PROCESS EQL LCURLY process_body RCURLY   {$4}
   | {[]}
   ;

schedtypedef:
    SCHEDULER EQL LCURLY process_body RCURLY {$4}
   | {[]}
  ;

process_body:
   process_var_decl SEMICOLON {[$1]}
   | process_var_decl SEMICOLON process_body {$1::$3}
  ;

process_var_decl:
  x_setpos requires type_expr IDENT
      { Mast.mkVARDECL($2,$3,$4,Module_misc.getpos 4) }
  | x_setpos requires TIMER IDENT
      {Mast.mkVARDECL($2,B.TIMER(B.NOT_PERSISTENT),$4,Module_misc.getpos 4)}
  | x_setpos requires type_expr SYSTEM IDENT
      {Mast.mkVARDECL($2,$3,$5,Module_misc.getpos 5)}
   ;

requires:
  REQUIRES {Mast.REQUIRES}
  | {Mast.NOT_REQUIRES}
  ;

   
const_var_decl:
  x_setpos bossa_type_expr IDENT
      {Mast.mkVARDECL(Mast.NOT_REQUIRES,$2,$3,Module_misc.getpos 3)}
   ;

constdef:
   x_setpos CONST const_var_decl EQL expr SEMICOLON constdef
        {Mast.mkCONSTDEF($3,$5,Module_misc.getpos 7) :: $7 }
  |     {[]}
   ;

parameter_types:
  param_type_list        {$1}
  |                      {[]}
  ;

param_type_list:
  type_expr                             {[$1]}
  | TIMER                               {[B.TIMER(B.NOT_PERSISTENT)]}
  | type_expr COMMA param_type_list     {$1::$3}
  | TIMER COMMA param_type_list         {B.TIMER(B.NOT_PERSISTENT)::$3}

globaldef:
  x_setpos non_proc_type_expr IDENT EQL expr SEMICOLON globaldef
      {let ln = Module_misc.getpos 7 in
      Mast.VL(Mast.mkVALDEF(Mast.mkVARDECL(Mast.NOT_REQUIRES,$2,$3,ln),$5,ln)) :: $7 }
  | x_setpos REQUIRES non_proc_type_expr IDENT SEMICOLON globaldef
      {let ln = Module_misc.getpos 6 in
      Mast.VL(Mast.mkSYSDEF(Mast.mkVARDECL(Mast.REQUIRES,$3,$4,ln),ln))::$6}
  | x_setpos non_proc_type_expr SYSTEM IDENT SEMICOLON globaldef
      {let ln = Module_misc.getpos 6 in
      Mast.VL(Mast.mkSYSDEF(Mast.mkSYSVARDECL(Mast.NOT_REQUIRES,$2,$4,ln),ln)) :: $6}
  | x_setpos non_proc_type_expr IDENT LPAREN parameter_types RPAREN SEMICOLON
      globaldef
      {Module_progtypes.add_fn_name $3;
	Mast.FN(Mast.mkFUNDECL($2,$3,$5,Module_misc.getpos 8)) :: $8 }
  | x_setpos VOID IDENT LPAREN parameter_types RPAREN SEMICOLON
      globaldef
      {Module_progtypes.add_fn_name $3;
	Mast.FN(Mast.mkFUNDECL(B.VOID,$3,$5,Module_misc.getpos 8)) :: $8 }
  | x_setpos requires TIMER IDENT SEMICOLON globaldef
        {let loc = Module_misc.getpos 6 in
	  Mast.VL(Mast.mkSYSDEF(Mast.mkVARDECL($2,
					       B.TIMER(B.NOT_PERSISTENT),$4,
					       loc),
				loc))
	  :: $6
        }
  |   {[]}
  ;

valdef:
  x_setpos type_expr IDENT EQL expr SEMICOLON valdef
	{let ln = Module_misc.getpos 7 in
          Mast.mkVALDEF(Mast.mkVARDECL(Mast.NOT_REQUIRES,$2,$3,ln),$5,ln) :: $7 }
  |     {[]}
   ;

/* ------------------------------------------------------------------ */
/* ordering criteria for schedulers */

sd_orderdef:
   ORDERING LCURLY order_body RCURLY     {$3}
  |                                      {[]}
   ;

order_body:
  | crit_decl   		  {[$1]}
  | crit_decl COMMA order_body    {$1::$3}
  ;

crit_decl:
   x_setpos MODULE_NAME
      {let ln = Module_misc.getpos 2 in
      Mast.mkIMPORT($2,ln)}
  ;
  
key:
   KEY             {Mast.KEY(0,0)}
  |                {Mast.NOKEY}

crit_order:
   LOWEST          {Mast.LOWEST}
   | HIGHEST       {Mast.HIGHEST}

/* ------------------------------------------------------------------ */
/* ordering criteria for modules */

orderdef:
  ORDERING EQL LCURLY module_order_body RCURLY {$4}
  |                                      {[]}
  ;

module_order_body:
  | module_crit_decl   		  {[$1]}
  | module_crit_decl COMMA module_order_body    {$1::$3}
  ;

module_crit_decl:
   x_setpos key crit_order IDENT     
      {Mast.mkMOD_CRIT_ID($2,$3,$4,Module_misc.getpos 4)}
  | x_setpos key crit_order LPAREN expr QUESTION expr COLON expr RPAREN
      {Mast.mkMOD_CRIT_TEST($2,$3,$5,$7,$9,Module_misc.getpos 10)}
  ;


/* ------------------------------------------------------------------ */
/* admissibility criteria for schedulers */

sd_admitdef:
    ADMIT x_setpos LCURLY admit_body RCURLY
       {let ln = Module_misc.getpos 4 in
       Mast.mkIMPORT_ADMIT($4,ln)}
  | x_setpos                          {Mast.mkNO_ADMIT(Module_misc.getpos 1)}
   ;

admit_body:
  | MODULE_NAME                    {[$1]}
  | MODULE_NAME COMMA admit_body   {$1::$3}
  ;

/* ------------------------------------------------------------------ */
/* admissibility criteria for modules */

admission:
  x_setpos ADMIT 
    EQL LCURLY valdef adm_crit adm_attach_detach RCURLY
    { let (crit_param,crit) = $6 in
    let (attach,detach) = $7 in
    Some(Mast.mkMOD_ADMIT($5,crit_param,crit,attach,detach,
			  Module_misc.getpos 8)) }  
  | { None }

adm_crit:
  ADMIT_CRIT parameters EQL LCURLY expr RCURLY         { ($2,$5) }

adm_attach_detach:
  ADMIT_ATT proc_param EQL seq_stmt ADMIT_DET proc_param EQL seq_stmt
      { (Some($2,$4),Some($6,$8)) }
  |   { (None,None) }

proc_param:
  x_setpos LPAREN requires proc_type IDENT RPAREN
      { Mast.mkVARDECL($3,$4,$5,Module_misc.getpos 6) }

/* ------------------------------------------------------------------ */
/* trace */

tracedef:
   x_setpos TRACE INTEGER LCURLY trace_events trace_expressions trace_test
            RCURLY
       {match $5@$6@$7 with
	 [] -> Mast.NOTRACE
       | all -> Mast.mkTRACE($3, all, Module_misc.getpos 8)}
   |   {Mast.NOTRACE}

trace_events:
   EVENTS EQL LCURLY event_types RCURLY SEMICOLON
       {[Mast.TRACEEVENTS($4)]}
   | {[]}

trace_expressions:
   x_setpos EXPRESSIONS EQL LCURLY fields RCURLY SEMICOLON
       {[Mast.TRACEEXPS(Mast.mkTRACEEXP($5,Module_misc.getpos 7))]}
   | {[]}

trace_test:
   TEST EQL LCURLY expr RCURLY SEMICOLON
       {[Mast.TRACETEST($4)]}
   | {[]}

fields:
   IDENT                       {[$1]}
   | IDENT COMMA fields        {$1::$3}
   | STATE_NAME                {[$1]}
   | STATE_NAME COMMA fields   {$1::$3}

statedef:
   STATES EQL LCURLY state_body RCURLY{$4}
   ;

state_body:
   state_decl 		    {[$1]}
   | state_decl state_body  {$1::$2}
   ;

state_decl:
     x_setpos class_name IDENT COLON PROCESS SEMICOLON
      { B.check_proc "process";
        Module_progtypes.add_state_name $3;
        Mast.mkPROCESS($2,$3,Class_state.PUBLIC,Class_state.SHARED,
		       Module_misc.getpos 6)
      }
   | x_setpos class_name IDENT COLON queue_visibility SCHEDULER SEMICOLON
      { B.check_virt "scheduler";
        Module_progtypes.add_state_name $3;
        Mast.mkPROCESS($2,$3,$5,Class_state.SHARED,Module_misc.getpos 7)
      }
   | x_setpos class_name IDENT COLON queue_visibility queue_type
       QUEUE SEMICOLON
      { Module_progtypes.add_state_name $3;
        Mast.mkQUEUE($2,$6,$3,$5,Class_state.SHARED,Module_misc.getpos 8)
      }
   | x_setpos class_name IDENT SEMICOLON
      {Module_progtypes.add_state_name $3;
        Mast.mkQUEUE($2,Class_state.EMPTY,$3,Class_state.PRIVATE,
		     Class_state.SHARED,Module_misc.getpos 4)
      }
   ;

class_name:
   READY			{Class_state.READY}
   | RUNNING			{Class_state.RUNNING}
   | BLOCKED			{Objects.is_using_blocked();
				  match (!Objects.policy_blocked_types,
					 !Objects.schedtype) with
				    (true,B.PROCSCHED) ->
				      Class_state.KERNEL_BLOCKED
				  | _ ->
				      Class_state.BLOCKED}
   | TERMINATED			{Class_state.TERMINATED}
   ;

not_terminated_class_name:
   READY			{Class_state.READY}
   | RUNNING			{Class_state.RUNNING}
   | BLOCKED			{Objects.is_using_blocked();
				  match (!Objects.policy_blocked_types,
					 !Objects.schedtype) with
				    (true,B.PROCSCHED) ->
				      Class_state.KERNEL_BLOCKED
				  | _ ->
				      Class_state.BLOCKED}
   ;

queue_visibility:
   PUBLIC			{Class_state.PUBLIC}
   | PRIVATE			{Class_state.PRIVATE}
   |				{Class_state.PUBLIC}
   ;

queue_type:
  SELECT			{Class_state.SELECT(Class_state.NODEFAULT)}
  | SELECT FIFO			{Class_state.SELECT(Class_state.FIFODEFAULT)}
  | SELECT LIFO			{Class_state.SELECT(Class_state.LIFODEFAULT)}
  |                             {Class_state.FIFO}
  ;

handlerdef:
   HANDLER LPAREN EVENT IDENT RPAREN LCURLY event_body RCURLY
	{
            List.map
	    (function x -> (x (Mast.mkVARDECL(Mast.NOT_REQUIRES,B.EVENT,$4,-1))))
	    ($7)}
  | {[]};

event_body:
   event_decl           	{$1}
  | event_decl event_body       {$1@$2}
  ;

event_decl:
  x_setpos ON event_types_overrideable seq_stmt
		{let ln = Module_misc.getpos 4 in
		List.map
		  (function (nm,overrideable) ->
		    function param ->
		      Mast.mkEVENT(nm,param,overrideable,$4,ln))
		  $3 }
  ;

event_types:
  x_setpos event_type_
      {[$2 (Module_misc.getpos 2)]}
  | x_setpos event_type_ COMMA event_types
      {($2 (Module_misc.getpos 4)) :: $4}
;

event_types_overrideable:
  x_setpos event_type_ overrideable
      {[($2 (Module_misc.getpos 3), $3)]}
  | x_setpos event_type_ overrideable COMMA event_types_overrideable
      {($2 (Module_misc.getpos 5), $3) :: $5}
;

event_type_:
  event_name hevt_list          {
                                  function x ->
                                  let (l,s) = $2 in
                                  Mast.mkEVENT_NAME($1::l,s,x)}
  ;

event_name:
  IDENT                         {$1}
  | STATE_NAME                  { $1 }
  | PROCESS                     {"process"}
  | SYSTEM                      {"system"}
  | TIMER                       {"timer"}

hevt_list:
  DOT TIMES			{([],true)}
  | DOT event_name hevt_list    {let (l,s) = $3 in ($2::l,s)}
  |                             {([],false)}
  ;

overrideable:
  OVERRIDEABLE                  {Mast.OVERRIDEABLE}
  |                             {Mast.NOT_OVERRIDEABLE}

/* ------------------------------------------------------------------ */

aspectdef:
   TRANSITION LPAREN PROCESS IDENT RPAREN EQL LCURLY aspect_body RCURLY
	{
            B.check_proc "process";
            List.map
	    (function x -> (x (Mast.mkVARDECL(Mast.NOT_REQUIRES,B.PROCESS,$4,-1))))
	    ($8)}
  | TRANSITION LPAREN SCHEDULER IDENT RPAREN EQL LCURLY aspect_body RCURLY
	{
            B.check_virt "scheduler";
            List.map
	    (function x -> (x (Mast.mkVARDECL(Mast.NOT_REQUIRES,B.SCHEDULER,$4,-1))))
	    ($8)}
  | { [] }
   ;

aspect_body:
  one_aspect_body               {[$1]}
  | one_aspect_body aspect_body {$1 :: $2}

one_aspect_body:
  | x_setpos ON aspect_header seq_stmt {
    let ln = Module_misc.getpos 4 in
    let (src,dst) = $3 in
    (function param -> Mast.mkASPECT(src,dst,param,Mast.ON,$4,ln)) }
  | x_setpos BEFORE aspect_header seq_stmt {
    let ln = Module_misc.getpos 4 in
    let (src,dst) = $3 in
    (function param -> Mast.mkASPECT(src,dst,param,Mast.BEFORE,$4,ln)) }
  | x_setpos AFTER aspect_header seq_stmt {
    let ln = Module_misc.getpos 4 in
    let (src,dst) = $3 in
    (function param -> Mast.mkASPECT(src,dst,param,Mast.AFTER,$4,ln)) }

aspect_header:
   nowhere_class_state RARROW state_name_list 	{ ([$1],$3) }
   | nowhere_class_state COMMA asp_state_name_list RARROW state_name_list
       { ($1::$3,$5) }

asp_state_name_list :
   nowhere_class_state  		           { [$1] }
   | nowhere_class_state COMMA asp_state_name_list { $1::$3 }
   ;

state_name_list :
   class_state 				{ [$1] }
   | class_state COMMA state_name_list  { $1::$3 }
   ;

class_state :
   STATE_NAME                           { $1 }
   | class_name                         { Class_state.class2idc $1 }

nowhere_class_state :
   class_state                          { $1 }
   | NOWHERE                            { "NOWHERE" }

/* ------------------------------------------------------------------ */

interfacedef:
  INTERFACE EQL LCURLY interface_body RCURLY {$4}
  |{[]}
  ;

interface_body:
     fundef		        {[$1]}
   | fundef interface_body      {$1::$2}
   ;

fundef:
        x_setpos type_or_void IDENT int_parameters seq_stmt
	{Mast.mkFUNDEF($2,$3,$4,$5,Module_misc.getpos 5)}
  ;

/* ------------------------------------------------------------------ */
/* parameters of interface functions */

int_parameters:
 LPAREN int_param_lists_body RPAREN   {$2}
 | LPAREN RPAREN 	          {[]}
 ;

int_param_lists_body:
   int_param_var_decl 	                   {[$1]}
   | int_param_var_decl COMMA int_param_lists_body {$1::$3}
   ;

int_param_var_decl:
  x_setpos requires type_expr IDENT
      {Mast.mkVARDECL($2,$3,$4,Module_misc.getpos 4)}
   ;

/* ------------------------------------------------------------------ */
/* parameters of functions and admission criteria */

parameters:
 LPAREN param_lists_body RPAREN   {$2}
 | LPAREN RPAREN 	          {[]}
 ;

param_lists_body:
   param_var_decl 	                   {[$1]}
   | param_var_decl COMMA param_lists_body {$1::$3}
   ;

param_var_decl:
  x_setpos requires type_expr IDENT
      {Mast.mkVARDECL($2,$3,$4,Module_misc.getpos 4)}
   ;

/* ------------------------------------------------------------------ */
/* parameters of modules functions */

parameters_modules:
 LPAREN param_mod_lists_body RPAREN   {$2}
 | LPAREN RPAREN 	          {[]}
 ;

param_mod_lists_body:
   state_module {[$1]}
   | state_module COMMA param_mod_lists_body { $1 :: $3 }
   ;
   
state_module: 
   queue_parameter {$1}
   | proc_parameter {$1}
   ;
   
queue_parameter:
   x_setpos not_terminated_class_name queue_visibility unshared queue_type
       QUEUE IDENT 
     {Module_progtypes.add_state_name $7;
     Mast.mkQUEUE($2,$5,$7,$3,$4,Module_misc.getpos 7)
     }
   | x_setpos TERMINATED unshared IDENT 
     {Module_progtypes.add_state_name $4;
     Mast.mkQUEUE(Class_state.TERMINATED,Class_state.EMPTY,$4,
		  Class_state.PRIVATE,$3,Module_misc.getpos 4)
     }
   ;
   
proc_parameter:
   x_setpos not_terminated_class_name queue_visibility unshared proc_type
    IDENT 
     {Module_progtypes.add_state_name $6;
     Mast.mkPROCESS($2,$6,$3,$4,Module_misc.getpos 6)
     }
   ;

unshared:
   UNSHARED                      {Class_state.UNSHARED}
   |                             {Class_state.SHARED}

bossa_type_expr:
  base_type	 {$1}
  | proc_type	 {$1}
  | ENUM_TYPE 	 {B.ENUM(B.mkId($1))}
  | RANGE_TYPE 	 {B.RANGE(B.mkId($1))}
  ;

type_expr:
  bossa_type_expr         {$1}
  | SYSTEM STRUCT IDENT   {B.SYSTEM(B.STRUCT(B.mkId($3)))}
  ;
  
non_proc_type_expr:
  base_type	          {$1}
  | ENUM_TYPE             {B.ENUM(B.mkId($1))}
  | RANGE_TYPE 	          {B.RANGE(B.mkId($1))}
  | SYSTEM STRUCT IDENT   {B.SYSTEM(B.STRUCT(B.mkId($3)))}
  ;

base_type:
  INT				{B.INT}
  | BOOL			{B.BOOL}
  | TIME 			{B.TIME}
  | CYCLES 			{B.CYCLES}
  | PORT			{B.PORT}
  ;

proc_type:
  | PROCESS			{B.check_proc "process";B.PROCESS}
  | SCHEDULER			{B.check_virt "scheduler";B.SCHEDULER}
  ;

type_or_void:
  type_expr		{$1}
  | VOID		{B.VOID}
  ;

stmt:
  if_stmt    		{$1}
  | for_stmt		{$1}
  | return_stmt 	{$1}
  | switch_stmt		{$1}
  | seq_stmt   		{$1}
  | assign_stmt	SEMICOLON	{$1}
/*  | defer_stmt          {$1}*/
  | prim_stmt           {$1}
  | error_stmt          {$1}
  | break_stmt          {$1}
  | super_stmt          {$1}
  ;

if_stmt:
  x_setpos if_condition seq_stmt  
      {Mast.mkIF($2,$3,Mast.mkSEQ([],[],-1),Module_misc.getpos 3)}
  | x_setpos if_condition seq_stmt ELSE seq_stmt	
	{Mast.mkIF($2,$3,$5,Module_misc.getpos 5)}

;

if_condition:
  IF LPAREN expr RPAREN  {$3}
  ;

/* foreach increasing and decreasing is only supported when only one state
name (not even a class name that is inly implemented by one state) is given
*/
for_stmt:
  x_setpos FOREACH LPAREN IDENT IN state_name_list RPAREN seq_stmt 
	{Mast.mkFOR($4,$6,$8,Module_misc.getpos 8)}
  | x_setpos FOREACHINC LPAREN IDENT IN STATE_NAME RPAREN seq_stmt 
	{Mast.mkFORINC($4,$6,$8,Module_misc.getpos 8)}
  | x_setpos FOREACHDEC LPAREN IDENT IN STATE_NAME RPAREN seq_stmt 
	{Mast.mkFORDEC($4,$6,$8,Module_misc.getpos 8)}
  | x_setpos FOREACH LPAREN IDENT RPAREN seq_stmt 
	{Mast.mkFOR_ANY($4,$6,Module_misc.getpos 6)}
  ;

switch_stmt:
  x_setpos SWITCH loc_expr IN seq_case
      {Mast.mkSWITCH($3,$5,Module_misc.getpos 5)}
  ;

return_stmt:
  x_setpos RETURN expr SEMICOLON	{Mast.mkRETURN($3,Module_misc.getpos 4)}
  | x_setpos RETURN SEMICOLON		{Mast.mkRETURNVOID(Module_misc.getpos 3)}
  ;

error_stmt:
  x_setpos ERROR LPAREN STRING RPAREN SEMICOLON
      {Mast.mkERROR($4,Module_misc.getpos 6)}

break_stmt:
  x_setpos BREAK SEMICOLON
      {Mast.mkBREAK(Module_misc.getpos 3)}

super_stmt:
  x_setpos NEXT LPAREN RPAREN SEMICOLON
      {Mast.mkNEXT(Module_misc.getpos 5)}

seq_case:
  LCURLY case_stmt RCURLY	{$2}
  ;

case_stmt:
  x_setpos CASE state_lists COLON seq_stmt case_stmt
      {Mast.mkSEQ_CASE($3,$5,Module_misc.getpos 6)::$6}
  |   {[]}
  ;

state_lists:
  STATE_NAME		         {[$1]}
  | STATE_NAME COMMA state_lists {$1::$3}
  | class_name		         {[Class_state.class2idc($1)]}
  | class_name COMMA state_lists {(Class_state.class2idc($1))::$3}
  ;

field_ref:
  IDENT	        {$1}
  ;

seq_stmt:
   x_setpos LCURLY valdef s_body RCURLY	{Mast.mkSEQ($3,$4,Module_misc.getpos 5)}
   ;

s_body:
  stmt s_body   {$1::$2}
  |		{[]}
  ;

/*
defer_stmt:
   x_setpos DEFER LPAREN RPAREN SEMICOLON
      {B.check_virt "defer()";Mast.mkDEFER(Module_misc.getpos 5)}
*/

prim_stmt:
   x_setpos FN_NAME LPAREN arglist RPAREN SEMICOLON
      {Mast.mkPRIMSTMT($2,$4,Module_misc.getpos 6)}

assign_stmt:
  x_setpos assign_stmt_         {$2(Module_misc.getpos 2)}

assign_stmt_:
  loc_expr PLUS_PLUS
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.PLUS,Mast.mkINT(1,x),x)}
  | loc_expr MINUS_MINUS
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.MINUS,Mast.mkINT(1,x),x)}
  | loc_expr EQL expr {function x ->  Mast.mkASSIGN($1,$3,x)}
  | loc_expr PLUS_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.PLUS,$3,x)}
  | loc_expr MINUS_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.MINUS,$3,x)}
  | loc_expr TIMES_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.TIMES,$3,x)}
  | loc_expr DIV_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.DIV,$3,x)}
  | loc_expr MOD_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.MOD,$3,x)}
  | loc_expr BITAND_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.BITAND,$3,x)}
  | loc_expr BITOR_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.BITOR,$3,x)}
  | loc_expr LSHIFT_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.LSHIFT,$3,x)}
  | loc_expr RSHIFT_EQUAL expr
      {function x -> Mast.mkCOMPOUND_ASSIGN($1,Mast.RSHIFT,$3,x)}
/* This ought to allow an arbitrary expr to the left of the arrow, but
doing so causes parsing problems.  Fortunately it seems that only a loc_expr
or select() can be type correct. */
  | move_expr RARROW STATE_NAME
      {function x -> Mast.mkMOVE($1,$3,x)}
  | move_expr RARROW STATE_NAME DOT HEAD
      {function x -> Mast.mkMOVEHD($1,$3,x)}
  | move_expr RARROW STATE_NAME DOT TAIL
      {function x -> Mast.mkMOVETL($1,$3,x)}
  | move_expr RARROW FWD LPAREN RPAREN
      {match !B.schedtype with
        B.PROCSCHED ->
	  raise (Error.Error "forwardImmediate() only in a virtual scheduler")
      |	B.VIRTSCHED -> function x -> Mast.mkMOVEFWD($1,x)}
  | move_expr RARROW FWD LPAREN RPAREN DOT HEAD
      {match !B.schedtype with
        B.PROCSCHED ->
	  raise (Error.Error "forwardImmediate() only in a virtual scheduler")
      |	B.VIRTSCHED -> function x -> Mast.mkMOVEHDFWD($1,x)}
  | move_expr RARROW FWD LPAREN RPAREN DOT TAIL
      {match !B.schedtype with
        B.PROCSCHED ->
	  raise (Error.Error "forwardImmediate() only in a virtual scheduler")
      |	B.VIRTSCHED -> function x -> Mast.mkMOVETLFWD($1,x)}
  ;

expr:
  x_setpos expr_                {$2(Module_misc.getpos 2)}

expr_:   
   INTEGER      {function x -> Mast.mkINT($1,x)}
   | IDENT      {function x -> Mast.mkVAR($1,x)}
   | STATE_NAME {function x -> Mast.mkVAR($1,x)}
   /*| MODULE_NAME {function x -> Mast.mkVAR($1,x)}*/
   | TRUE	{function x -> Mast.mkBOOL(true,x)}
   | FALSE	{function x -> Mast.mkBOOL(false,x)}
   | PLUS expr_  {function x -> Mast.mkBINARY(Mast.PLUS,Mast.mkINT(0,x),$2 x,x)}
   | MINUS expr_ {function x -> Mast.mkBINARY(Mast.PLUS,Mast.mkINT(0,x),$2 x,x)}
   | TIMES expr_ {function x -> Mast.mkINDR($2 x, x)}
   | NOT expr_ {function x -> Mast.mkUNARY(Mast.NOT, $2 x, x)}
   | COMPLEMENT expr_  {function x -> Mast.mkUNARY(Mast.COMPLEMENT, $2 x, x)}
   | expr_ DOT field_ref  {function x -> Mast.mkFIELD($1 x,$3,x)}
   | SELECT LPAREN RPAREN {function x -> Mast.mkSELECT(x)}
   | FN_NAME LPAREN arglist RPAREN {function x -> Mast.mkPRIM($1,$3,x)}
   | EMPTY LPAREN STATE_NAME RPAREN {function x -> Mast.mkEMPTY($3,x)}
   | EMPTY LPAREN class_name RPAREN {function x -> Mast.mkEMPTYCLASS($3,x)}
   | SCHEDULEROF LPAREN expr RPAREN {function x -> Mast.mkSCHEDCHILD($3,x)}
   | SRCONSCHED LPAREN RPAREN {function x -> Mast.mkSRCONSCHED(x)}
   | MODULE_NAME DOT IDENT {B.check_module "ordering criteria";
                            function x -> Mast.mkFIELD_OC($1,$3,x)
                           }
   | expr_ PLUS expr_ {function x -> Mast.mkBINARY(Mast.PLUS,$1 x,$3 x,x)}
   | expr_ MINUS expr_ {function x -> Mast.mkBINARY(Mast.MINUS,$1 x,$3 x,x)}
   | expr_ TIMES expr_ {function x -> Mast.mkBINARY(Mast.TIMES,$1 x,$3 x,x)}
   | expr_ DIV expr_ {function x -> Mast.mkBINARY(Mast.DIV,$1 x,$3 x,x)}
   | expr_ MOD expr_ {function x -> Mast.mkBINARY(Mast.MOD,$1 x,$3 x,x)}
   | expr_ AND expr_ {function x -> Mast.mkBINARY(Mast.AND,$1 x,$3 x,x)}
   | expr_ OR expr_ {function x -> Mast.mkBINARY(Mast.OR,$1 x,$3 x,x)}
   | expr_ BITAND expr_ {function x -> Mast.mkBINARY(Mast.BITAND,$1 x,$3 x,x)}
   | expr_ BITOR expr_ {function x -> Mast.mkBINARY(Mast.BITOR,$1 x,$3 x,x)}
   | expr_ EQUAL_EQUAL expr_ {function x -> Mast.mkBINARY(Mast.EQ,$1 x,$3 x,x)}
   | expr_ NEQ expr_ {function x -> Mast.mkBINARY(Mast.NEQ,$1 x,$3 x,x)}
   | expr_ LT expr_ {function x -> Mast.mkBINARY(Mast.LT,$1 x,$3 x,x)}
   | expr_ GT expr_ {function x -> Mast.mkBINARY(Mast.GT,$1 x,$3 x,x)}
   | expr_ LEQ expr_ {function x -> Mast.mkBINARY(Mast.LEQ,$1 x,$3 x,x)}
   | expr_ GEQ expr_ {function x -> Mast.mkBINARY(Mast.GEQ,$1 x,$3 x,x)}
   | expr_ LSHIFT expr_ {function x -> Mast.mkBINARY(Mast.LSHIFT,$1 x,$3 x,x)}
   | expr_ RSHIFT expr_ {function x -> Mast.mkBINARY(Mast.RSHIFT,$1 x,$3 x,x)}
   | expr_ IN STATE_NAME {function x -> Mast.mkIN($1 x, $3, x)}
   | expr_ IN class_name {function x -> Mast.mkINCLASS($1 x, $3, x)}
   | LPAREN expr RPAREN {function x -> $2}
    ;

arglist:
   arglist1                      {$1}
   |                             {[]}
   ;

arglist1:
   expr                          {[$1]}
   | expr COMMA arglist1         {$1::$3}

loc_expr:
   x_setpos loc_expr_             {$2(Module_misc.getpos 2)}

loc_expr_:
  IDENT         	{function x -> Mast.mkVAR($1,x)}
  | STATE_NAME   	{function x -> Mast.mkVAR($1,x)}
  | loc_expr_ DOT IDENT {function x -> Mast.mkFIELD($1 x,$3,x)}
  ;

move_expr:
  x_setpos SELECT LPAREN RPAREN         {Mast.mkSELECT(Module_misc.getpos 4)}
  | x_setpos SCHEDULEROF LPAREN expr RPAREN
      {Mast.mkSCHEDCHILD($4,Module_misc.getpos 5)}
  | loc_expr                              {$1}


/* helper production. Purpose: sideeffect: set lineno to get position
where expr starts */
x_setpos:
	  {Stack.push (!Module_misc.lineno, !Module_misc.cr_index) Module_misc.pos_stack;}  
	  ;
x_setvirt:
	  {B.schedtype := B.VIRTSCHED}  
	  ;

x_setproc:
	  {B.schedtype := B.PROCSCHED}  
	  ;

x_setmodule:
          {B.schedmodule := B.MODULE}
         
