(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* --------------------- Program-defined types --------------------- *)

module B = Objects

module HashString =
struct
	type t = string
	let equal (s1 : t) (s2 : t) = s1 = s2
	let hash (s : t) = Hashtbl.hash s
end
module StringHashtbl = Hashtbl.Make(HashString)

type idtype =
    ENUM_TYPE | RANGE_TYPE | IDENT | FN_NAME | STATE_NAME | MODULE_NAME

let lexicon = StringHashtbl.create 211
let init_lexicon _ =
  StringHashtbl.clear lexicon;
  (* copy from user *)
  StringHashtbl.add lexicon "get_user_int" FN_NAME;
  (* primitive time-related functions *)
  StringHashtbl.add lexicon "now" FN_NAME;
  StringHashtbl.add lexicon "start_absolute_timer" FN_NAME;
  StringHashtbl.add lexicon "start_relative_timer" FN_NAME;
  StringHashtbl.add lexicon "stop_timer" FN_NAME;
  StringHashtbl.add lexicon "time_to_ticks" FN_NAME;
  StringHashtbl.add lexicon "ticks_to_time" FN_NAME;
  if !B.high_res
  then
    begin
      StringHashtbl.add lexicon "make_time" FN_NAME;
      StringHashtbl.add lexicon "make_cycle_time" FN_NAME;
      StringHashtbl.add lexicon "make_cycles" FN_NAME;
      StringHashtbl.add lexicon "time_to_seconds" FN_NAME;
      StringHashtbl.add lexicon "time_to_nanoseconds" FN_NAME;
      StringHashtbl.add lexicon "time_to_jiffies" FN_NAME;
      StringHashtbl.add lexicon "time_to_subjiffies" FN_NAME
    end

let add_enum_type name =
  StringHashtbl.add lexicon name ENUM_TYPE

let add_range_type name =
  StringHashtbl.add lexicon name RANGE_TYPE

let add_fn_name name =
  StringHashtbl.add lexicon name FN_NAME

let add_state_name name =
  StringHashtbl.add lexicon name STATE_NAME

let add_module_name name =
  StringHashtbl.add lexicon name MODULE_NAME

let scan_ident id =
  try StringHashtbl.find lexicon id
  with Not_found -> IDENT
