(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)


(* ---------------------------- parser error ---------------------------- *)

open Lexing
open Format

type t = { loc_start: int; loc_end: int }
      
let symbol_rloc () =
    { loc_start = Parsing.symbol_start();
      loc_end   = Parsing.symbol_end() }
    
let rhs_loc n =
  { loc_start = Parsing.rhs_start n;
    loc_end   = Parsing.rhs_end n }
	  
    
let print ppf loc =
  fprintf ppf "File \"%s\", line %i" 
    (Misc.get_filename()) (Misc.get_linenum loc.loc_start);
  fprintf ppf ", characters %i" 
    (loc.loc_start - (Misc.get_linebeg loc.loc_start));
  fprintf ppf "-%i:\n" 
    (loc.loc_end - (Misc.get_linebeg loc.loc_end))



(* Report an error *)

let report_error lexbuf ppf exn =
  let report ppf = function
    | Bossa_lexer.Error(err, start, stop) ->
	print ppf {loc_start = start; loc_end = stop};
	Bossa_lexer.report_error ppf err

(*    | Syntaxerr.Error(loc, err) ->
	Location.print ppf loc;
	Syntaxerr.report_error ppf err*)

    | Parsing.Parse_error ->
	print ppf { loc_start = Lexing.lexeme_start lexbuf; 
			     loc_end   = Lexing.lexeme_end lexbuf};
	fprintf ppf "Syntax error"

    | x -> fprintf ppf "@]"; raise x in
  
  fprintf ppf "@[%a@]@." report exn
