(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* ----------------------------- Classes ---------------------------- *)

type classname = RUNNING | READY | BLOCKED | TERMINATED | NOWHERE
  | POLICY_BLOCKED | KERNEL_BLOCKED | KERNEL_POLICY_BLOCKED
val all_classes : bool (* true if called from event types and VS needs
			  terminated *) -> classname list
val all_classes_nowhere : bool (* as above *) -> classname list
val class2c : classname -> string
val safe_class2c : classname -> string
val classic_class2c : classname -> string
val class2idc : classname -> string
val class2state : classname -> string
val id2class : Objects.identifier -> classname
val less_than : classname * classname -> bool

(* ----------------------- State information ------------------------ *)

type state_info =
    STATE of Objects.identifier * classname * state_type * visibility
and state_type  = PROC | QUEUE of queue_order
and queue_order = FIFO | SELECT of sort_default | EMPTY
and sort_default = LIFODEFAULT | FIFODEFAULT | NODEFAULT
and visibility = PUBLIC | PRIVATE
and share_state = SHARED | UNSHARED
val state2c : state_info -> string
val state2class : state_info -> classname

(* ----------------------------- States ----------------------------- *)

val state_env : (Objects.identifier * state_info) list ref
val selected : state_info option ref
val selected_size : int ref
val concretize : Objects.identifier -> state_info list
val concretize_class : classname -> state_info list
val all_states : unit -> state_info list

(* ------------------------- States and classes --------------------- *)

val valid_sources : Objects.identifier -> classname -> state_info list
