(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

type entry = Class_state.classname * Vobjects.vtype * Vobjects.modified

type type_rule =
    Objects.isquasi *
      (entry list * Class_state.classname list) list * entry list list

val init_events : Ast.event_name list ref

val system_events : unit -> Ast.event_name list

val automaton :
    unit -> (int * Ast.event_name * Objects.intinfo * int) list
val init_automaton :
    unit -> (int * Ast.event_name * Objects.intinfo * int) list
val automaton_start : int

val init_event_env : unit -> unit

val set_os : string -> unit

val all_event_names : Ast.event_name list ref

(* --------------------- top-level event names ---------------------- *)

val flatten_specd_name :
    Ast.event_name list -> Ast.event_name -> Ast.event_name

(* the names that the policy should define *)
val top_level_event_names : Ast.event_name list ref
val top_level_events : unit -> Ast.event_name list list

(* filled in later, in verification, when handlers are available *)
val specialized_events : ((Ast.event_name list * Ast.event_name) list) ref

val top_level_event_names_loop :
    (Ast.event_name list -> Ast.event_name ->
      Objects.identifier list list -> bool -> Objects.attribute -> 'a) ->
	(Ast.event_name -> 'b) ->
	  Ast.event_name list list ->
	    ('a list * 'b) list


val event_specifications :
    unit -> (Ast.event_name * type_rule list) list

val vs_event_specifications :
    unit -> (Ast.event_name * type_rule list) list

val interface_specifications :
    unit -> (Objects.identifier * type_rule list) list

val vs_interface_specifications :
    unit -> (Objects.identifier * type_rule list) list
